using System;

namespace TecnoComm {
    public interface device {

        // List devices of this type that are still available
        // RETURN: list of devices
        string[] listDevices();

        // Open a device for communication
        // device: identify device to open: should be the index (0 based) of the device
        //         in the list obtained by listDevices
        // baud  : baud rate required for communication; other serial parameters are
        //         fixed: 8N1.
        // RETURN: true on success
        bool open(int device, uint baud);

        // Close device and releases it
        // RETURN: true on success
        bool close();

        // Sends a byte on the communication channel
        // b     : byte to send
        // RETURN: true on success
        bool putByte(byte b);

        // Get a byte from the communication channel
        // b     : if a byte is available, it is stored here
        // RETURN: true if a byte was retrieved
        bool getByte(ref byte b);

        // Empty input and output buffers
        // RETURN: true on success
        bool flush();

        // Synchronizes communication speed
        // RETURN: true synchronization succeeded
        bool synchronize();

        // Signal if the PC is ready (should map to DTR)
        // on: set to true to signal ready
        // RETURN: true on success
        bool signal(bool on);

    }

    public class error : System.ApplicationException {
        public error(string message)
            :
            base(message) {
        }
    }

    public class errorNOPORT : error {
        public errorNOPORT()
            :
            base("Comm port not initialized") {
        }
    }

    public class errorTIMEOUT : error {
        public errorTIMEOUT()
            :
            base("Timeout during communication") {
        }
    }

    public class errorFORMAT : error {
        public errorFORMAT()
            :
            base("Bad message format") {
        }
    }

    public class errorBADTYPE : error {
        public errorBADTYPE()
            :
            base("Bad message type") {
        }
    }

    public class packet {
        public const uint MAX_CONTENT_SIZE = packetDirectory.MAX_CONTENT_SIZE;

        public byte dest;     // Unit this packet is sent to
        public byte source;   // Unit sending this packet
        public byte type;     // Packet type (application specific)
        public UInt16 size;     // Size of packet content
        public UInt16 sequence; // Sequence number of packet (application specific)
        public byte[] content = new byte[MAX_CONTENT_SIZE]; // Payload (content is application specific)

        // Master unit identifier
        public const byte MASTER_ID = 0xFF;
        // Packet is to be received by everyone
        public const byte ALL_ID = 0x00;

        public packet() {
            dest = ALL_ID;
            source = MASTER_ID;
        }

        public void copyTo(packet pkt) {
            pkt.dest = dest;
            pkt.source = source;
            pkt.type = type;
            pkt.size = size;
            pkt.sequence = sequence;
            for (uint i = 0; i < MAX_CONTENT_SIZE; i++) pkt.content[i] = content[i];
        }


        public void contentWriteByte(byte b, int offset) {
            content[offset] = b;
        }

        public void contentWriteBytes(byte[] bytes, uint num, int offset) {
            for (uint i = 0; i < num; i++)
                content[offset + i] = num < bytes.Length ? bytes[i] : (byte)0;
        }

        public void contentWriteBytes(byte[] bytes, int offset) {
            for (uint i = 0; i < bytes.Length; i++) content[offset + i] = bytes[i];
        }

        public void contentWriteUInt16(UInt16 u16, int offset) {
            byte[] val = BitConverter.GetBytes(u16);
            contentWriteBytes(val, offset);
        }

        public void contentWriteUInt32(UInt32 u32, int offset) {
            byte[] val = BitConverter.GetBytes(u32);
            contentWriteBytes(val, offset);
        }

        public void contentWriteFloat(float f, int offset) {
            byte[] val = BitConverter.GetBytes(f);
            contentWriteBytes(val, offset);
        }

        public void contentWriteString(string s,uint maxLen, int offset) {
            for (int i = 0; i < maxLen; i++) {
                content[offset + i] = i < s.Length ? (byte)s[i] : (byte)0;
            }
        }

        public void contentWriteDateTime(DateTime time,int offset) {
            DateTime epoch = new DateTime(2000, 1, 1);
            UInt32 v = (UInt32)((time.Ticks - epoch.Ticks) / 10000000);
            contentWriteUInt32(v, offset);
        }

        public byte contentReadByte(int offset) {
            return content[offset];
        }

        public byte[] contentReadBytes(uint num, int offset) {
            byte[] ret = new byte[num];

            for (uint i = 0; i < num; i++) ret[i] = content[offset + i];
            return ret;
        }

        public UInt16 contentReadUInt16(int offset) {
            return BitConverter.ToUInt16(content, offset);
        }

        public UInt32 contentReadUInt32(int offset) {
            return BitConverter.ToUInt32(content, offset);
        }

        public float contentReadFloat(int offset) {
            return BitConverter.ToSingle(content, offset);
        }

        public string contentReadString(uint maxLen, int offset) {
            System.Text.StringBuilder sb = new System.Text.StringBuilder(16);
            for (int i = 0; i < maxLen; i++) {
                byte b = content[offset + i];
                if (b == 0) break;
                sb.Append((char)b);
            }
            return sb.ToString();
        }

        public DateTime contentReadDateTime(int offset) {
            DateTime dt = new DateTime(2000, 1, 1);
            return dt.AddSeconds(contentReadUInt32(offset));
        }


    }

    public class communication {
        /* NOTE ON BYTE PACKING IN PACKETS
        Bytes 0xFF are used as packet delimiters; so 0xFF is send before each packet
        in order to reset receiver state machine.
        Bytes 0xFE are used as escape sequence start inside packets so that the
        next byte is used to encode a serie of up to four 0xFE, 0xFF bytes.
        The byte following 0xFE is used to encode up to four bytes as follows.

        x1 x0 y1 y0 z1 z0 t1 t0
        |  |  |  |  |  |  |  |
        |  |  |  |  |  |  +--+--- Fourth byte
        |  |  |  |  +--+--------- Third byte
        |  |  +--+--------------- Second byte
        +--+--------------------- First byte

        Each couple of bits has the following meaning:

        00 No byte encoded
        01 0xFE byte encoded
        10 0xFF byte encoded
        11 FORBIDDEN

        For example 00100100 would encode two bytes 0xFF, 0xFE
        */

        // Identifiers for addressing packets

        // Physical medium access
        private device io;

        // Checksum of packet being sent
        private uint sndChecksum1;
        private uint sndChecksum2;

        // Escape status of packet being sent
        private byte sndEscape;

        // Status of the receive state machine
        private uint rcvState;
        private uint rcvSubState;

        // Packet being received
        private packet rcvPacket;

        // Checksum of packet being received
        private uint rcvChecksum1;
        private uint rcvChecksum2;

        // Escape status of packet being received
        private byte rcvEscape;

        // Creator
        public communication() {
            rcvState = 0;
            rcvSubState = 0;
            rcvPacket = new packet();
            io = null;
        }

        // Destructor
        ~communication() {
            freeReaderPort();
        }

        // Free communication port; called by destructor
        public void freeReaderPort() {
            if (io != null)
            {
                io.close();
                io = null;
            }
        }

        // Send a raw byte to communication port
        // RETURN: true if OK, false on error
        private bool putRawByte(byte b) {
            return io.putByte(b);
        }

        // Send bytes to communication port inside a packet (handling escape sequences)
        // bytes: pointer to bytes to send
        // count: number of bytes to send
        // RETURN: true if OK, false on error
        private bool putBytes(ref byte[] bytes, uint count) {
            byte c;
            uint i = 0;

            while (count-- != 0) {
                c = bytes[i++];
                sndChecksum1 += c; if (sndChecksum1 >= 255) sndChecksum1 -= 255;
                sndChecksum2 += sndChecksum1; if (sndChecksum2 >= 255) sndChecksum2 -= 255;
                if (c >= 0xFE) {
                    if (sndEscape == 0) {
                        if (!putRawByte(0xFE)) return false;
                    }
                    sndEscape <<= 2;
                    sndEscape |= (byte)((c & 0x03) - 1); // Encode 0xFE anf 0xFF as 01b or 10b
                    if ((sndEscape & 0xC0) != 0) {
                        if (!putRawByte(sndEscape)) return false;
                        sndEscape = 0;
                    }
                }
                else {
                    if (sndEscape != 0) {
                        if (!putRawByte(sndEscape)) return false;
                        sndEscape = 0;
                    }
                    if (!putRawByte(c)) return false;
                }
            }
            return true;
        }

        // Send byte to communication port inside a packet (handling escape sequences)
        // b     : byte to send
        // RETURN: true if OK, false on error
        private bool putByte(byte b) {
            byte[] b1 = new byte[1];

            b1[0] = b;
            return putBytes(ref b1, 1);
        }

        // Send Uint16 to communication port inside a packet, low byte first
        // (handling escape sequences)
        // u     : Uint16 to send
        // RETURN: true if OK, false on error        
        private bool putUint16(UInt16 u) {
            byte[] b2 = new byte[2];

            b2[0] = (byte)u;
            b2[1] = (byte)(u >> 8);
            return putBytes(ref b2, 2);
        }

        private const uint stWaitStart = 0;
        private const uint stWaitDest = 1;
        private const uint stWaitSource = 2;
        private const uint stWaitType = 3;
        private const uint stWaitSize1 = 4;
        private const uint stWaitSize2 = 5;
        private const uint stWaitSeq1 = 6;
        private const uint stWaitSeq2 = 7;
        private const uint stWaitData = 8;
        private const uint stWaitSum1 = 9;
        private const uint stWaitSum2 = 10;

        bool synched = false;

        // Set port to use for communication
        // io    : port to use (set to null to just free com port)
        public void setReaderPort(device io) {
            freeReaderPort();
            synched = false;
            if (io != null) {
                this.io = io;
                rcvState = stWaitStart;
            }
            else throw new errorNOPORT();
        }

        // Get reader port
        // RETURN: reference to reader port
        device getReaderPort() {
            return io;
        }

        // Get a packet if available: note that packet is NOT used for temporary
        // storage so it does not have to be the same pointer when polling
        // packet: place for storing received packet
        // RETURN: true if a packet was read
        public bool getPacket(packet pkt) {
            byte c = 0;

            if (io == null) throw new errorNOPORT();

            for (; ; ) {
                if (rcvEscape == 0) {
                    if (!io.getByte(ref c)) return false;
                    if (c == 0xFF) {
                        rcvState = stWaitDest; rcvSubState = 0;
                        rcvChecksum1 = rcvChecksum2 = 0;
                        continue;
                    }
                    if (c == 0xFE) { rcvState |= 0x40; continue; }
                    if ((rcvState & 0x40) != 0) { rcvState &= 0x3F; rcvEscape = c; continue; }
                }
                else {
                    if ((rcvEscape & 0xC0) == 0x40) c = 0xFE;
                    else if ((rcvEscape & 0xC0) == 0x80) c = 0xFF;
                    else { rcvEscape <<= 2; continue; }
                    rcvEscape <<= 2;
                }

                rcvChecksum1 += c; if (rcvChecksum1 >= 255) rcvChecksum1 -= 255;
                rcvChecksum2 += rcvChecksum1; if (rcvChecksum2 >= 255) rcvChecksum2 -= 255;

                switch (rcvState) {
                    case stWaitDest:
                        if (c == packet.MASTER_ID) { rcvPacket.dest = c; rcvState = stWaitSource; }
                        else rcvState = stWaitStart;
                        break;

                    case stWaitSource: rcvPacket.source = c; rcvState = stWaitType; break;

                    case stWaitType: rcvPacket.type = c; rcvState = stWaitSize1; break;

                    case stWaitSize1: rcvPacket.size = c; rcvState = stWaitSize2; break;

                    case stWaitSize2:
                        rcvPacket.size += (UInt16)(((UInt16)c) << 8);
                        if (rcvPacket.size <= packet.MAX_CONTENT_SIZE) rcvState = stWaitSeq1;
                        else rcvState = stWaitStart;
                        break;

                    case stWaitSeq1:
                        rcvPacket.sequence = c; rcvState = stWaitSeq2; break;

                    case stWaitSeq2:
                        rcvPacket.sequence += (UInt16)(((UInt16)c) << 8);
                        if (rcvPacket.size!=0) {
                            rcvState = stWaitData;
                            rcvSubState = 0;
                        }
                        else rcvState = stWaitSum1;
                        break;

                    case stWaitData:
                        rcvPacket.content[rcvSubState++] = c;
                        if (rcvSubState == rcvPacket.size) rcvState = stWaitSum1;
                        break;

                    case stWaitSum1: rcvState = stWaitSum2; break;

                    case stWaitSum2:
                        rcvState = stWaitStart;
                        if ((rcvChecksum1 == 0xFF || rcvChecksum1 == 0x00) &&
                            (rcvChecksum2 == 0xFF || rcvChecksum2 == 0x00)) {
                            rcvPacket.copyTo(pkt);
                            return true;
                        }
                        rcvState = stWaitStart;
                        break;

                    default: rcvState = stWaitStart;
                        break;

                }
            }
        }

        // Send packet to communication port
        // packet: packet to send
        public void putPacket(packet pkt) {
            uint tw;
            byte tb;

            if (io == null) throw new errorNOPORT();

            sndEscape = 0;
            sndChecksum1 = sndChecksum2 = 0;
            if (!putRawByte(0xFF)) throw new errorNOPORT();
            if (!putByte(pkt.dest)) throw new errorNOPORT();
            if (!putByte(pkt.source)) throw new errorNOPORT();
            if (!putByte(pkt.type)) throw new errorNOPORT();
            if (!putUint16(pkt.size)) throw new errorNOPORT();
            if (!putUint16(pkt.sequence)) throw new errorNOPORT();
            if (!putBytes(ref pkt.content, pkt.size)) throw new errorNOPORT();

            tw = sndChecksum1 + sndChecksum2; if (tw >= 255) tw -= 255;
            tb = (byte)sndChecksum2;

            if (!putByte((byte)(255 - tw))) throw new errorNOPORT();
            if (!putByte((byte)tb)) throw new errorNOPORT();

            if (sndEscape != 0) {
                if (!putRawByte(sndEscape)) throw new errorNOPORT();
            }
        }

        // Send a packet and polls for answer
        // pktSnd : packet to send
        // pktRcv : place for storing received packet
        // timeout: max time to wait in milliseconds
        public void exchangePacket(packet pktSnd, packet pktRcv, uint timeout) {
            int start;
           
            if (io == null) throw new errorNOPORT();

            io.flush();

            if (!synched) io.synchronize();

            start = System.Environment.TickCount;
            putPacket(pktSnd);
            do {
                System.Threading.Thread.Sleep(1);
                if (getPacket(pktRcv)) {
                    synched = true;
                    return;
                }
            } while (System.Environment.TickCount - start <= timeout);
            synched = false;
            throw new errorTIMEOUT();
        }

    }
}
