using System;
using System.Collections.Generic;
using System.Text;

namespace TecnoComm {
    public class packetDirectory {

        // Packets that are common to all transducers
        public const byte NET_UNIT = 0x00;
        public const byte NET_CHANNEL = 0x01;
        public const byte NET_READ = 0x02;

        // Maximum content of a single packet
        public const uint MAX_CONTENT_SIZE = 512+2;
    }

    public class packetNET_UNIT : packet {

        public packetNET_UNIT()
            : base() {
            type = packetDirectory.NET_UNIT;
        }

        private void checkGet() {
            if (type != packetDirectory.NET_UNIT) throw new errorBADTYPE();
            if (size < SIZE) throw new errorFORMAT();
        }

        private void checkSet() {
            size = SIZE;
        }

        // The following fields are only in answer message
        //unsigned char identity[8];
        //unsigned      model;      // Model of intelligent transducer
        //unsigned      numberOfChannels;
        //unsigned long calibration;// Calibration date (in seconds since 1/1/2000)
        //unsigned long expiry;     // Expiration date (in seconds since 1/1/2000)

        private const int OFS_IDENTITY = 0;

        public byte[] identity {
            get { checkGet(); return contentReadBytes(8, OFS_IDENTITY); }
            set {
                checkSet();
                contentWriteBytes(value, 8, OFS_IDENTITY);
            }
        }

        private const int OFS_MODEL = OFS_IDENTITY + 8;

        public UInt16 model {
            get { checkGet(); return contentReadUInt16(OFS_MODEL); }
            set { checkSet(); contentWriteUInt16(value, OFS_MODEL); }
        }

        private const int OFS_NUMBEROFCHANNELS = OFS_MODEL + 2;

        public UInt16 numberOfChannels {
            get { checkGet(); return contentReadUInt16(OFS_NUMBEROFCHANNELS); }
            set { checkSet(); contentWriteUInt16(value, OFS_NUMBEROFCHANNELS); }
        }

        private const int OFS_CALIBRATION = OFS_NUMBEROFCHANNELS + 2;

        public DateTime calibration {
            get { checkGet(); return contentReadDateTime(OFS_CALIBRATION); }
            set { checkSet(); contentWriteDateTime(value, OFS_CALIBRATION); }
        }

        private const int OFS_EXPIRY = OFS_CALIBRATION + 4;

        public DateTime expiry {
            get { checkGet(); return contentReadDateTime(OFS_EXPIRY); }
            set { checkSet(); contentWriteDateTime(value, OFS_EXPIRY); }
        }

        private const int SIZE = OFS_EXPIRY + 4;
    }

    public class packetNET_CHANNEL : packet {

        public packetNET_CHANNEL()
            : base() {
            type = packetDirectory.NET_CHANNEL;
        }

        private void checkGet() {
            if (type != packetDirectory.NET_CHANNEL) throw new errorBADTYPE();
            if (size < SIZE) throw new errorFORMAT();
        }

        private void checkSet() {
            size = SIZE;
        }

        // Constants for measure field below describing how to interpret unit of measure fields

        // 0: Unit is described by the product ofsi base units, plus radians and steradians,
        // raised to the powers recorded in fields 2-10.
        public const byte MEASURE_TYPE_SI = 0;

        //  1: Unit is U/U, where U is described by the product of SI base units, plus radians
        // and steradians, raised to the powers recorded in fields 2-10.
        public const byte MEASURE_TYPE_UU = 1;

        // 2: Unit is loglO(U), where U is described by the product of SI base units, plus radians
        // and steradians, raised to the powers recorded in fields 2-10.
        public const byte MEASURE_TYPE_LOG10U = 2;

        // 3: Unit is log10(U/U),whereU is described by the product of SI base units, plus radians
        // and steradians, raised to the powers recorded in fields 2-10.
        public const byte MEASURE_TYPE_LOG10UU = 3;

        // 4: The associated quantity is digital data (e.g., a bit vector) and has no unit.
        // Fields 2-10 shall be set to 128. The "digital data" type applies to data that represents counting,
        // such as the output of an ADC, or that represents a state, such as the current positions
        // of a bank of switches.
        public const byte MEASURE_TYPE_DIGITAL = 4;

        // 5: The associated physical quantity is represented by values on an arbitrary scale
        // (e.g., hardness). Fields 2-10 are reserved, and shall be set to 128
        public const byte MEASURE_TYPE_ARBITRARY = 5;

        // 6-255: Reserved

        private byte ENCODE_EXPONENT(double exponent) { return (byte)(128.0 + 2.0 * exponent); }

        private double DECODE_EXPONENT(byte exponent) { return ((double)exponent - 128.0) / 2.0; }


        private const int OFS_CHANNEL = 0;

        public UInt16 channel {
            get { checkGet(); return contentReadUInt16(OFS_CHANNEL); }
            set { size = 2; contentWriteUInt16(value, OFS_CHANNEL); }
        }

        private const int OFS_TYPE = OFS_CHANNEL + 2;

        public UInt16 transducer_type {
            get { checkGet(); return contentReadUInt16(OFS_TYPE); }
            set { checkSet(); contentWriteUInt16(value, OFS_TYPE); }
        }

        private const int OFS_SUPPLY = OFS_TYPE + 2;

        public UInt16 supply {
            get { checkGet(); return contentReadUInt16(OFS_SUPPLY); }
            set { checkSet(); contentWriteUInt16(value, OFS_SUPPLY); }
        }

        private const int OFS_UNIT = OFS_SUPPLY + 2;

        public string unit {
            get { checkGet(); return contentReadString(16, OFS_UNIT); }
            set { checkSet(); contentWriteString(value, 16, OFS_UNIT); }
        }

        private const int OFS_MEASURE = OFS_UNIT + 16;

        public byte measure {
            get { checkGet(); return contentReadByte(OFS_MEASURE); }
            set { checkSet(); contentWriteByte(value, OFS_MEASURE); }
        }

        public byte[] measureExponents {
            get { checkGet(); return contentReadBytes(8, OFS_RADIANS); }
            set { checkSet(); 
                contentWriteBytes(value, 9, OFS_RADIANS);
                for (int i = value.Length; i < 9; i++) contentWriteByte(128, OFS_RADIANS + i);
            }
        }

        private const int OFS_RADIANS = OFS_MEASURE + 1;

        public byte radians {
            get { checkGet(); return contentReadByte(OFS_RADIANS); }
            set { checkSet(); contentWriteByte(value, OFS_RADIANS); }
        }

        private const int OFS_STERADIANS = OFS_RADIANS + 1;

        public byte steradians {
            get { checkGet(); return contentReadByte(OFS_STERADIANS); }
            set { checkSet(); contentWriteByte(value, OFS_STERADIANS); }
        }

        private const int OFS_METERS = OFS_STERADIANS + 1;

        public byte meters {
            get { checkGet(); return contentReadByte(OFS_METERS); }
            set { checkSet(); contentWriteByte(value, OFS_METERS); }
        }

        private const int OFS_KILOGRAMS = OFS_METERS + 1;

        public byte kilograms {
            get { checkGet(); return contentReadByte(OFS_KILOGRAMS); }
            set { checkSet(); contentWriteByte(value, OFS_KILOGRAMS); }
        }

        private const int OFS_SECONDS = OFS_KILOGRAMS + 1;

        public byte seconds {
            get { checkGet(); return contentReadByte(OFS_SECONDS); }
            set { checkSet(); contentWriteByte(value, OFS_SECONDS); }
        }

        private const int OFS_AMPERES = OFS_SECONDS + 1;

        public byte amperes {
            get { checkGet(); return contentReadByte(OFS_AMPERES); }
            set { checkSet(); contentWriteByte(value, OFS_AMPERES); }
        }

        private const int OFS_KELVINS = OFS_AMPERES + 1;

        public byte kelvins {
            get { checkGet(); return contentReadByte(OFS_KELVINS); }
            set { checkSet(); contentWriteByte(value, OFS_KELVINS); }
        }

        private const int OFS_MOLES = OFS_KELVINS + 1;

        public byte moles {
            get { checkGet(); return contentReadByte(OFS_MOLES); }
            set { checkSet(); contentWriteByte(value, OFS_MOLES); }
        }

        private const int OFS_CANDELAS = OFS_MOLES + 1;

        public byte candelas {
            get { checkGet(); return contentReadByte(OFS_CANDELAS); }
            set { checkSet(); contentWriteByte(value, OFS_CANDELAS); }
        }

        private const int SIZE = OFS_CANDELAS + 1;
    }

    public class packetNET_READ : packet {

        // No commamd: just sending message to query reading status
        public const UInt16 CHANNEL_COMMAND_NONE = 0;

        // Start reading channel
        public const UInt16 CHANNEL_COMMAND_START = 1;

        // Channel has been read and value is being returned
        public const UInt16 CHANNEL_READ_OK = 0x0000;
        // Channel is in overflow condition
        public const UInt16 CHANNEL_OVERFLOW = 0x0001;
        // Channel is in underflow condition
        public const UInt16 CHANNEL_UNDERFLOW = 0x0002;
        // Readout not ready yet, but wait
        public const UInt16 CHANNEL_WAIT = 0xFE00;
        // Failure reading channel (the low byte will contain additional information)
        public const UInt16 CHANNEL_FAILURE = 0xFF00;

        public packetNET_READ()
            : base() {
            type = packetDirectory.NET_READ;
        }

        private void checkGet() {
            if (type != packetDirectory.NET_READ) throw new errorBADTYPE();
            if (size < SIZE) throw new errorFORMAT();
        }

        private void checkSet() {
            size = SIZE;
        }

        private const int OFS_CHANNEL = 0;

        public UInt16 channel {
            get { checkGet(); return contentReadUInt16(OFS_CHANNEL); }
            set { size = 4; contentWriteUInt16(value, OFS_CHANNEL); }
        }

        private const int OFS_COMMAND = OFS_CHANNEL + 2;

        public UInt16 command {
            get { checkGet(); return contentReadUInt16(OFS_COMMAND); }
            set { size = 4; contentWriteUInt16(value, OFS_COMMAND); }
        }

        private const int OFS_VALUE = OFS_COMMAND + 2;

        public float value {
            get { checkGet(); return contentReadFloat(OFS_VALUE); }
            set { size = 4; contentWriteFloat(value, OFS_VALUE); }
        }

        protected const int OFS_ERROR = OFS_VALUE + 4;

        public UInt16 error {
            get { checkGet(); return contentReadUInt16(OFS_ERROR); }
            set { size = 4; contentWriteUInt16(value, OFS_ERROR); }
        }

        private const int SIZE = OFS_ERROR + 2;
    }

    
}