﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Runtime.InteropServices;

namespace TecnoComm {

    public class deviceRS485 : device {

        private System.IO.Ports.SerialPort port;
        private string[] portNames = new string[0];

        uint baudRate;

        [DllImport("Kernel32.dll")]
        extern private static UInt32 GetTickCount();

        // List devices of this type that are still available
        // RETURN: list of devices
        public string[] listDevices() {
            portNames = System.IO.Ports.SerialPort.GetPortNames();
            for (int i = 0; i < portNames.Length - 1; i++) {
                for (int j = i + 1; j < portNames.Length; j++) {
                    if (portNames[i].CompareTo(portNames[j]) > 0) {
                        string sw = portNames[i];
                        portNames[i] = portNames[j];
                        portNames[j] = sw;
                    }
                }
            }
            return portNames;
        }

        // Has a character been sent and echo not received yet?
        bool waitEcho = false;
        // Character that was sent
        byte waitEchoWhat;

        // Open a device for communication
        // device: identify device to open: should be the index (0 based) of the device
        //         in the list obtained by listDevices
        // baud  : baud rate required for communication; other serial parameters are
        //         fixed: 8N1.
        // RETURN: true on success
        public bool open(int device, uint baud) {
            UInt32 start;
            close();
            waitEcho = false;
            start = GetTickCount();
            while (GetTickCount() - start < 100) ;
            try {
                port = new System.IO.Ports.SerialPort(portNames[device], (int)baud);
                port.Open();
                port.ReadTimeout = 0;
                port.DtrEnable = false;
                port.RtsEnable = false;
                baudRate = baud;
                start = GetTickCount();
                while (GetTickCount() - start < 100) ;
                return true;
            }
            catch (System.Exception) {
            }
            return false;
        }

        // Close device and releases it
        // RETURN: true on success
        public bool close() {
            if (port == null) return true;
            if (port.IsOpen) {
                port.DtrEnable = false;
                port.Close();
            }
            port = null;
            return true;
        }

        // Object used to send a single byte to system routines that want
        // an array
        byte[] oneByteArray = new byte[1];

        // Check that echo is coming back
        // RETURN: true if echo is detected matching last character sent
        private bool checkEcho() {
            if (!waitEcho) return true;
            waitEcho = false;

            UInt32 start = GetTickCount();
            UInt32 wait = (UInt32)(10 * 9 * 1000 / port.BaudRate + 200);
            do {
                try {
                    if (port.Read(oneByteArray, 0, 1) == 1) {
                        return oneByteArray[0] == waitEchoWhat;
                    }
                }
                catch (System.TimeoutException) { }
                catch (System.IO.IOException) { }
            } while (GetTickCount() - start <= wait);

            return false;
        }

        // Sends a byte on the communication channel
        // b     : byte to send
        // RETURN: true on success
        public bool putByte(byte b) {
            if (port.DtrEnable == false) {
                port.DtrEnable = true;
                port.RtsEnable = true;
                System.Threading.Thread.Sleep(1);
            }
            if (port == null) return false;
            if (!checkEcho()) return false;
            oneByteArray[0] = b;
            port.Write(oneByteArray, 0, 1);
            waitEchoWhat = b;
            waitEcho = true;
            return true;
        }

        // Input buffer
        private const uint inBufferSize = 64;
        private byte[] inBuffer = new byte[inBufferSize];
        private uint inBufferNext;
        private uint inBufferLength;

        // Fill input buffer
        private void fillInBuffer() {
            inBufferNext = 0;
            try {
                inBufferLength = (uint)port.Read(inBuffer, 0, (int)inBufferSize);
            }
            catch (System.TimeoutException) {
                inBufferLength = 0;
            }
            catch (System.IO.IOException) {
                inBufferLength = 0;
            }
        }

        // Get a byte from the communication channel
        // b     : if a byte is available, it is stored here
        // RETURN: true if a byte was retrieved
        public bool getByte(ref byte b) {
            if (port.DtrEnable == true) {
                port.DtrEnable = false;
                port.RtsEnable = false;
                System.Threading.Thread.Sleep(1);
            }
            if (port == null) return false;
            if (!checkEcho()) return false;
            if (inBufferNext >= inBufferLength) {
                fillInBuffer();
                if (inBufferNext >= inBufferLength) return false;
            }
            b = inBuffer[inBufferNext++];
            return true;
        }

        // Empty input and output buffers
        // RETURN: true on success
        public bool flush() {
            if (port == null) return false;
            port.DiscardOutBuffer();
            port.DiscardInBuffer();
            inBufferNext = inBufferLength = 0;
            waitEcho = false;
            return true;
        }

        // Synchronizes communication speed
        // RETURN: true synchronization succeeded
        public bool synchronize() {
            /*
            port.DtrEnable = true;
            port.RtsEnable = true;
            System.Threading.Thread.Sleep(1);
            if (port == null) return false;
            flush();
            port.BaudRate = (int)(baudRate / 4);
            if (!putByte(0x00)) {
                port.BaudRate = (int)baudRate;
                return false;
            }
            bool done = checkEcho();
            port.BaudRate = (int)baudRate;
            System.Threading.Thread.Sleep(1);            
            port.DtrEnable = false;
            port.RtsEnable = false;
            return done;
            */
            return true;
        }

        // Signal if the PC is ready (should map to DTR)
        // on: set to true to signal ready
        public bool signal(bool on) {
            if (port == null) return false;
            port.DtrEnable = on;
            port.RtsEnable = on;
            return true;
        }

    }
}
